#include <bm.h>

#include "ext2fs/ext2_fs.h"
#include "ext2fs/ext2fs.h"

#include "ext2_fs.h"

/*
 * For checking structure magic numbers...
 */

#define EXT2_CHECK_MAGIC(struct, code) \
	  if ((struct)->magic != (code)) return (code)

struct blk_private_data {
	int	magic;
	int	align;
	ext2_loff_t offset;
	struct struct_io_stats io_stats;
	block_t *blk;
};

#define IS_ALIGNED(n, align) ((((uintptr_t) n) & \
			       ((uintptr_t) ((align)-1))) == 0)

static errcode_t unix_get_stats(io_channel channel, io_stats *stats)
{
	errcode_t	retval = 0;

	struct blk_private_data *data;

	EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
	data = (struct blk_private_data *) channel->private_data;
	EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_UNIX_IO_CHANNEL);

	if (stats) {
		*stats = &data->io_stats;
	}

	return retval;
}

/*
 * Here are the raw I/O functions
 */
static errcode_t raw_read_blk(io_channel channel,
			      struct blk_private_data *data,
			      unsigned long long block,
			      int count, void *bufv)
{
	errcode_t	retval;
	ssize_t		size;
	ext2_loff_t	location;
	ssize_t		actual = 0;
	char		*buf = bufv;

	size = (count < 0) ? -count : (ext2_loff_t) count * channel->block_size;
	location = ((ext2_loff_t) block * channel->block_size) + data->offset;

	data->io_stats.bytes_read += size;

	if (channel->align && !IS_ALIGNED(location, channel->align)) {
		/*
		 * The location which we're trying to read isn't aligned,
		 * so we need to do this the hard way...
		 */

		off_t aligned_blk = location / channel->align;
		int offset = location % channel->align;
		int need = channel->align - offset;

		actual = data->blk->pread(data->blk, buf, (size_t)channel->align, (off_t)aligned_blk);
		if (actual <= 0)
		{
			retval = -EIO;
			goto error_out;
		}
		
		memcpy(buf, buf+offset, need);
		buf += need;
		size -= need;
	}

	actual = data->blk->pread(data->blk, buf, (size_t)size, (off_t)location);
	if (actual == size)
		return 0;
	else if (actual >= 0)
	{
		memset(buf+actual, 0, size-actual);
		retval = EXT2_ET_SHORT_READ;
	} else
		retval = -EIO;

error_out:
	if (channel->read_error)
		retval = (channel->read_error)(channel, block, count, buf,
					       size, actual, retval);
	return retval;
}

static errcode_t raw_write_blk(io_channel channel,
			       struct blk_private_data *data,
			       unsigned long long block,
			       int count, const void *bufv)
{
	errcode_t	retval;
	ssize_t		size;
	ext2_loff_t	location;
	ssize_t		actual = 0;
	const char	*buf = bufv;

	size = (count < 0) ? -count : (ext2_loff_t) count * channel->block_size;
	location = ((ext2_loff_t) block * channel->block_size) + data->offset;

	data->io_stats.bytes_written += size;

	if (channel->align && !IS_ALIGNED(location, channel->align)) {
		/*
		 * The location which we're trying to write isn't aligned,
		 * so we need to do this the hard way...
		 */
		char tmp[channel->align];
		off_t aligned_blk = location / channel->align;
		int offset = location % channel->align;
		int need = channel->align - offset;

		actual = data->blk->pread(data->blk, tmp, (size_t)channel->align, (off_t)aligned_blk);
		if (actual != channel->align)
		{
			retval = -EIO;
			goto error_out;
		}

		memcpy(tmp+offset, buf, need);

		actual = data->blk->pwrite(data->blk, tmp, (size_t)channel->align, (off_t)aligned_blk);
		if (actual != channel->align)
		{
			retval = -EIO;
			goto error_out;
		}

		buf += need;
		size -= need;
	}

	actual = data->blk->pwrite(data->blk, buf, (size_t)size, (off_t)location);
	if (actual == size)
		return 0;
	else if (actual >= 0)
		retval = EXT2_ET_SHORT_READ;
	else
		retval = -EIO;

error_out:
	if (channel->write_error)
		retval = (channel->write_error)(channel, block, count, buf,
						size, actual, retval);
	return retval;
}

static errcode_t unix_open_channel(block_t *blk,
				   io_channel *channel,
				   io_manager io_mgr)
{
	io_channel	io = NULL;
	struct blk_private_data *data = NULL;
	errcode_t	retval;
	uint32_t sector_size;

	if (blk->ioctl(blk, IOCTL_GET_SECTOR_SIZE, &sector_size))
	{
		retval = -EIO;
		goto cleanup;
	}

	retval = ext2fs_get_mem(sizeof(struct struct_io_channel), &io);
	if (retval)
		goto cleanup;
	memset(io, 0, sizeof(struct struct_io_channel));
	io->magic = EXT2_ET_MAGIC_IO_CHANNEL;
	retval = ext2fs_get_mem(sizeof(struct blk_private_data), &data);
	if (retval)
		goto cleanup;

	io->manager = io_mgr;
	io->name = "BD I/O";
	io->private_data = data;
	io->align = sector_size;
	io->block_size = 1024;
	io->read_error = 0;
	io->write_error = 0;
	io->refcount = 1;
	io->flags = CHANNEL_FLAGS_BLOCK_DEVICE;

	memset(data, 0, sizeof(struct blk_private_data));
	data->magic = EXT2_ET_MAGIC_UNIX_IO_CHANNEL;
	data->io_stats.num_fields = 2;
	data->blk = blk;

	*channel = io;
	return 0;

cleanup:
	if (data)
		ext2fs_free_mem(&data);
	if (io)
		ext2fs_free_mem(&io);
	return retval;
}

static errcode_t unix_open(const char *name, int flags, io_channel *channel)
{
	block_t *blk = ((struct fake_name *)name)->blk;
	if (blk == NULL)
		return EXT2_ET_BAD_DEVICE_NAME;
	return unix_open_channel(blk, channel, unix_io_manager);
}

static errcode_t unix_close(io_channel channel)
{
	struct blk_private_data *data;
	errcode_t	retval = 0;

	EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
	data = (struct blk_private_data *) channel->private_data;
	EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_UNIX_IO_CHANNEL);

	if (--channel->refcount > 0)
		return 0;

	ext2fs_free_mem(&channel->private_data);
	ext2fs_free_mem(&channel);
	return retval;
}

static errcode_t unix_set_blksize(io_channel channel, int blksize)
{
	struct blk_private_data *data;
	errcode_t		retval = 0;

	EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
	data = (struct blk_private_data *) channel->private_data;
	EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_UNIX_IO_CHANNEL);

	channel->block_size = blksize;
	return retval;
}

static errcode_t unix_read_blk64(io_channel channel, unsigned long long block,
			       int count, void *buf)
{
	struct blk_private_data *data;

	EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
	data = (struct blk_private_data *) channel->private_data;
	EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_UNIX_IO_CHANNEL);

	return raw_read_blk(channel, data, block, count, buf);
}

static errcode_t unix_read_blk(io_channel channel, unsigned long block,
			       int count, void *buf)
{
	return unix_read_blk64(channel, block, count, buf);
}

static errcode_t unix_write_blk64(io_channel channel, unsigned long long block,
				int count, const void *buf)
{
	struct blk_private_data *data;

	EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
	data = (struct blk_private_data *) channel->private_data;
	EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_UNIX_IO_CHANNEL);

	return raw_write_blk(channel, data, block, count, buf);
}

static errcode_t unix_cache_readahead(io_channel channel,
				      unsigned long long block,
				      unsigned long long count)
{
	return EXT2_ET_OP_NOT_SUPPORTED;
}

static errcode_t unix_write_blk(io_channel channel, unsigned long block,
				int count, const void *buf)
{
	return unix_write_blk64(channel, block, count, buf);
}

static errcode_t unix_write_byte(io_channel channel, unsigned long offset,
				 int size, const void *buf)
{
	return EXT2_ET_UNIMPLEMENTED;
}

/*
 * Flush data buffers to disk.
 */
static errcode_t unix_flush(io_channel channel)
{
	return 0;
}

static errcode_t unix_set_option(io_channel channel, const char *option,
				 const char *arg)
{
	struct blk_private_data *data;
	unsigned long long tmp;
	char *end;

	EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
	data = (struct blk_private_data *) channel->private_data;
	EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_UNIX_IO_CHANNEL);

	if (!strcmp(option, "offset")) {
		if (!arg)
			return EXT2_ET_INVALID_ARGUMENT;

		tmp = strtoull(arg, &end, 0);
		if (*end)
			return EXT2_ET_INVALID_ARGUMENT;
		data->offset = tmp;
		if (data->offset < 0)
			return EXT2_ET_INVALID_ARGUMENT;
		return 0;
	}
	if (!strcmp(option, "cache")) {
		return 0;
	}
	return EXT2_ET_INVALID_ARGUMENT;
}

static errcode_t unix_discard(io_channel channel, unsigned long long block,
			      unsigned long long count)
{
	return EXT2_ET_UNIMPLEMENTED;
}

static errcode_t unix_zeroout(io_channel channel, unsigned long long block,
			      unsigned long long count)
{
	return EXT2_ET_UNIMPLEMENTED;
}

static struct struct_io_manager struct_unix_manager = {
	.magic		= EXT2_ET_MAGIC_IO_MANAGER,
	.name		= "BD I/O Manager",
	.open		= unix_open,
	.close		= unix_close,
	.set_blksize	= unix_set_blksize,
	.read_blk	= unix_read_blk,
	.write_blk	= unix_write_blk,
	.flush		= unix_flush,
	.write_byte	= unix_write_byte,
	.set_option	= unix_set_option,
	.get_stats	= unix_get_stats,
	.read_blk64	= unix_read_blk64,
	.write_blk64	= unix_write_blk64,
	.discard	= unix_discard,
	.cache_readahead	= unix_cache_readahead,
	.zeroout	= unix_zeroout,
};

io_manager unix_io_manager = &struct_unix_manager;
