/*
 * EXT FS support
 * 
 * Copyright (c) 2021 Ant-ON
 *
 */

#include <sys/param.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>

#include <fs.h>
#include <bm.h>
#include "ext2fs/ext2fs.h"
#include "ext2fs/ext2_fs.h"
#include "ext2fs/ext2_err.h"
#include "ext2_fs.h"

#include "log.h"

static int ext2_error(errcode_t err);

struct ext2_dir_iter
{
	void *ctx;
	ext2_filsys e2fs;
	 _filter filter;
};

struct ext2_rd_struct
{
	ext2_ino_t	parent;
	int empty;
};

struct ext2_upd_dotdot 
{
	ext2_ino_t new_dotdot;
};

static ssize_t ext2_f_read(void *vfd, void *buf, size_t len)
{
	fd_t *fd = (fd_t *)vfd;
	ext2_file_t file = (ext2_file_t)fd->data;

	unsigned int rcnt = 0;
	errcode_t ret = ext2fs_file_read(file, buf, len, &rcnt);
	if (rcnt > 0)
		return rcnt;
	else
		return ret;
}

static ssize_t ext2_f_write(void *vfd, const void *buf, size_t len)
{
	fd_t *fd = (fd_t *)vfd;
	ext2_file_t file = (ext2_file_t)fd->data;

	unsigned int wcnt = 0;
	errcode_t ret = ext2fs_file_write(file, buf, len, &wcnt);
	if (wcnt > 0)
		return wcnt;
	else
		return ext2_error(ret);
}

static off_t ext2_f_seek(void *vfd, off_t offset, int dir)
{
	fd_t *fd = (fd_t *)vfd;
	ext2_file_t file = (ext2_file_t)fd->data;

	__u64 off = 0;
	ext2fs_file_llseek(file, offset, dir, &off);
	return off;
}

static void ext2_f_close(void *vfd)
{
	fd_t *fd = (fd_t *)vfd;
	if (!fd->data)
		return;

	ext2_file_t file = (ext2_file_t)fd->data;
	fd->data = NULL;
	ext2fs_file_flush(file);
	ext2fs_flush(file->fs);
	ext2fs_file_close(file);
	//free(fd); - free in FS.c
}

/*
 * Extended fields will fit into an inode if the filesystem was formatted
 * with large inodes (-I 256 or larger) and there are not currently any EAs
 * consuming all of the available space. For new inodes we always reserve
 * enough space for the kernel's known extended fields, but for inodes
 * created with an old kernel this might not have been the case. None of
 * the extended inode fields is critical for correct filesystem operation.
 * This macro checks if a certain field fits in the inode. Note that
 * inode-size = GOOD_OLD_INODE_SIZE + i_extra_isize
 */
#define EXT4_FITS_IN_INODE(ext4_inode, field)		\
	((offsetof(typeof(*ext4_inode), field) +	\
	  sizeof((ext4_inode)->field))			\
	 <= ((size_t) EXT2_GOOD_OLD_INODE_SIZE +		\
	    (ext4_inode)->i_extra_isize))		\

static inline __u32 ext4_encode_extra_time(const struct timespec *time)
{
	__u32 extra = sizeof(time->tv_sec) > 4 ?
			((time->tv_sec - (__s32)time->tv_sec) >> 32) &
			EXT4_EPOCH_MASK : 0;
	return extra | (time->tv_nsec << EXT4_EPOCH_BITS);
}

#define EXT4_INODE_SET_XTIME(xtime, timespec, raw_inode)		       \
do {									       \
	(raw_inode)->xtime = (timespec)->tv_sec;			       \
	if (EXT4_FITS_IN_INODE(raw_inode, xtime ## _extra))		       \
		(raw_inode)->xtime ## _extra =				       \
				ext4_encode_extra_time(timespec);	       \
} while (0)

#define EXT4_EINODE_SET_XTIME(xtime, timespec, raw_inode)		       \
do {									       \
	if (EXT4_FITS_IN_INODE(raw_inode, xtime))			       \
		(raw_inode)->xtime = (timespec)->tv_sec;		       \
	if (EXT4_FITS_IN_INODE(raw_inode, xtime ## _extra))		       \
		(raw_inode)->xtime ## _extra =				       \
				ext4_encode_extra_time(timespec);	       \
} while (0)

static void init_times(struct ext2_inode_large *inode)
{
	struct timespec now;

	clock_gettime(CLOCK_REALTIME, &now);
	EXT4_INODE_SET_XTIME(i_atime, &now, inode);
	EXT4_INODE_SET_XTIME(i_ctime, &now, inode);
	EXT4_INODE_SET_XTIME(i_mtime, &now, inode);
	EXT4_EINODE_SET_XTIME(i_crtime, &now, inode);
}

static int ext2_can_allocate(ext2_filsys e2fs, blk64_t num)
{
	blk64_t reserved;

	if (num > ext2fs_blocks_count(e2fs->super))
		return 0;

	/*
	 * Different meaning for r_blocks -- libext2fs has bugs where the FS
	 * can get corrupted if it totally runs out of blocks.  Avoid this
	 * by refusing to allocate any of the reserve blocks to anybody.
	 */
	reserved = ext2fs_r_blocks_count(e2fs->super);
	if (reserved == 0)
		reserved = ext2fs_blocks_count(e2fs->super) / 10;
	return ext2fs_free_blocks_count(e2fs->super) > reserved + num;
}

static int ext2_file_type(unsigned int mode)
{
	if (LINUX_S_ISREG(mode))
		return EXT2_FT_REG_FILE;

	if (LINUX_S_ISDIR(mode))
		return EXT2_FT_DIR;

	if (LINUX_S_ISCHR(mode))
		return EXT2_FT_CHRDEV;

	if (LINUX_S_ISBLK(mode))
		return EXT2_FT_BLKDEV;

	if (LINUX_S_ISLNK(mode))
		return EXT2_FT_SYMLINK;

	if (LINUX_S_ISFIFO(mode))
		return EXT2_FT_FIFO;

	if (LINUX_S_ISSOCK(mode))
		return EXT2_FT_SOCK;

	return 0;
}

static ext2_ino_t ext2_create(ext2_filsys e2fs, const char *path, int mode)
{
	ext2_ino_t parent, child = 0;
	struct ext2_inode_large parent_inode, inode;
	char *temp_path, *node_name, a;
	int err;

	if (!ext2_can_allocate(e2fs, 1))
		return 0;

	temp_path = strdup(path);
	if (!temp_path)
		return 0;

	node_name = strrchr(temp_path, '/');
	if (!node_name)
		goto error_out;
	node_name++;
	a = *node_name;
	*node_name = 0;
	
	err = ext2fs_namei(e2fs, EXT2_ROOT_INO, EXT2_ROOT_INO, temp_path, &parent);
	if (err || parent == 0)
		goto error_out;

	err = ext2fs_read_inode_full(e2fs, parent, (struct ext2_inode *)&parent_inode, sizeof(parent_inode));
	if (err)
		goto error_out;
	if (!LINUX_S_ISDIR(parent_inode.i_mode))
		goto error_out;

	err = ext2fs_new_inode(e2fs, parent, mode, 0, &child);
	if (err || child == 0)
		goto error_out;

	// creating ino name in dir
	*node_name = a;
	err = ext2fs_link(e2fs, parent, node_name, child, ext2_file_type(mode));
	if (err == EXT2_ET_DIR_NO_SPACE)
	{
		err = ext2fs_expand_dir(e2fs, parent);
		if (!err)
			err = ext2fs_link(e2fs, parent, node_name, child, ext2_file_type(mode));
	}
	if (err)
	{
		child = 0;
		goto error_out;
	}

	memset(&inode, 0, sizeof(inode));
	inode.i_mode = mode;
	inode.i_links_count = 1;
	inode.i_extra_isize = sizeof(struct ext2_inode_large) - EXT2_GOOD_OLD_INODE_SIZE;
    inode.i_uid = parent_inode.i_uid;
    inode.osd2.linux2.l_i_uid_high = parent_inode.osd2.linux2.l_i_uid_high;
    inode.i_gid = parent_inode.i_gid;
    inode.osd2.linux2.l_i_gid_high = parent_inode.osd2.linux2.l_i_gid_high;

	if (ext2fs_has_feature_extents(e2fs->super))
	{
		ext2_extent_handle_t handle;

		inode.i_flags &= ~EXT4_EXTENTS_FL;
		err = ext2fs_extent_open2(e2fs, child, (struct ext2_inode *)&inode, &handle);
		if (err)
		{
			child = 0;
			goto error_out;
		}
		ext2fs_extent_free(handle);
	}

	err = ext2fs_write_new_inode(e2fs, child, (struct ext2_inode *)&inode);
	if (err)
	{
		child = 0;
		goto error_out;
	}

	init_times(&inode);
	inode.i_generation = inode.i_atime + 1;
	err = ext2fs_write_inode_full(e2fs, child, (struct ext2_inode *)&inode, sizeof(inode));
	if (err)
	{
		child = 0;
		goto error_out;
	}

	ext2fs_inode_alloc_stats2(e2fs, child, 1, 0);

error_out:
	free(temp_path);
	return child;
}

static void* ext2_fs_fopen(void *vfs, const char *path, int flags)
{
	fs_t *fs = (fs_t *)vfs;
	ext2_filsys e2fs = (ext2_filsys)fs->data;
	ext2_ino_t ino = 0;
	ext2_file_t file = NULL;
	int err;

	err = ext2fs_namei(e2fs, EXT2_ROOT_INO, EXT2_ROOT_INO, path, &ino);
	if (err || ino == 0)
	{
		if (flags & O_CREAT)
			ino = ext2_create(e2fs, path, 0755 | LINUX_S_IFREG);
		if (ino == 0)
			return NULL;
	} else
	{
		struct ext2_inode inode;
		
		err = ext2fs_read_inode(e2fs, ino, &inode);
		if (err)
			return NULL;
		if (!LINUX_S_ISREG(inode.i_mode))
			return NULL;
	}

	if (ext2fs_file_open(e2fs, ino, (flags&O_WRONLY)?EXT2_FILE_WRITE:0, &file) || !file)
		return NULL;

	fd_t *fd = malloc(sizeof(fd_t));
	fd->read = ext2_f_read;
	fd->write = ext2_f_write;
	fd->seek = ext2_f_seek;
	fd->close = ext2_f_close;
	fd->data = file;
	return fd;
}

static int ext2_fs_mkdir(void *vfs, const char *path)
{
	fs_t *fs = (fs_t *)vfs;
	ext2_filsys e2fs = (ext2_filsys)fs->data;
	char *temp_path = NULL, *node_name, a;
	ext2_ino_t parent, child;
	struct ext2_inode_large parent_inode, inode;
	char *block;
	blk64_t blk;
	errcode_t err = 0;

	if (!ext2_can_allocate(e2fs, 1)) {
		err = -ENOSPC;
		goto out;
	}

	temp_path = strdup(path);
	if (!temp_path) {
		err = -ENOMEM;
		goto out;
	}
	node_name = strrchr(temp_path, '/');
	if (!node_name) {
		err = -ENOMEM;
		goto out;
	}
	node_name++;
	a = *node_name;
	*node_name = 0;
	
	err = ext2fs_namei(e2fs, EXT2_ROOT_INO, EXT2_ROOT_INO, temp_path, &parent);
	if (err)
		goto out;

	err = ext2fs_read_inode_full(e2fs, parent, (struct ext2_inode *)&parent_inode, sizeof(parent_inode));
	if (err)
		goto out;

	*node_name = a;

	err = ext2fs_mkdir(e2fs, parent, 0, node_name);
	if (err == EXT2_ET_DIR_NO_SPACE)
	{
		err = ext2fs_expand_dir(e2fs, parent);
		if (!err)
			err = ext2fs_mkdir(e2fs, parent, 0, node_name);
	}
	if (err)
		goto out;

	/* Still have to update the uid/gid of the dir */
	err = ext2fs_namei(e2fs, EXT2_ROOT_INO, EXT2_ROOT_INO, temp_path, &child);
	if (err)
		goto out;

	memset(&inode, 0, sizeof(inode));
	err = ext2fs_read_inode_full(e2fs, child, (struct ext2_inode *)&inode, sizeof(inode));
	if (err)
		goto out;

	inode.i_mode = LINUX_S_IFDIR | 0755 | (parent_inode.i_mode & S_ISGID);
    inode.i_uid = parent_inode.i_uid;
    inode.osd2.linux2.l_i_uid_high = parent_inode.osd2.linux2.l_i_uid_high;
    inode.i_gid = parent_inode.i_gid;
    inode.osd2.linux2.l_i_gid_high = parent_inode.osd2.linux2.l_i_gid_high;
    inode.i_generation = time(NULL);

	err = ext2fs_write_inode_full(e2fs, child, (struct ext2_inode *)&inode, sizeof(inode));
	if (err)
		goto out;

	/* Rewrite the directory block checksum, having set i_generation */
	if ((inode.i_flags & EXT4_INLINE_DATA_FL) ||
	    !ext2fs_has_feature_metadata_csum(e2fs->super))
		goto out;

	err = ext2fs_new_dir_block(e2fs, child, parent, &block);
	if (err)
		goto out;

	err = ext2fs_bmap2(e2fs, child, (struct ext2_inode *)&inode, NULL, 0, 0, NULL, &blk);
	if (err)
		goto out2;
	
	err = ext2fs_write_dir_block4(e2fs, blk, block, 0, child);

out2:
	ext2fs_free_mem(&block);

out:
	if (temp_path)
		free(temp_path);
	return ext2_error(err);
}

static errcode_t ext2_remove_inode(ext2_filsys fs, ext2_ino_t ino)
{
	errcode_t err;
	struct ext2_inode_large inode;

	memset(&inode, 0, sizeof(inode));
	err = ext2fs_read_inode_full(fs, ino, (struct ext2_inode *)&inode, sizeof(inode));
	if (err)
		goto out;

	switch (inode.i_links_count) 
	{
	case 0:
		return 0; /* XXX: already done? */
	case 1:
		inode.i_links_count--;
		inode.i_dtime = fs->now ? fs->now : time(0);
		break;
	default:
		inode.i_links_count--;
	}

	if (inode.i_links_count)
		goto write_out;

	/* Nobody holds this file; free its blocks! */
	err = ext2fs_free_ext_attr(fs, ino, &inode);
	if (err)
		goto write_out;

	if (ext2fs_inode_has_valid_blocks2(fs, (struct ext2_inode *)&inode)) {
		err = ext2fs_punch(fs, ino, (struct ext2_inode *)&inode, NULL, 0, ~0ULL);
		if (err)
			goto write_out;
	}

	ext2fs_inode_alloc_stats2(fs, ino, -1, LINUX_S_ISDIR(inode.i_mode));

write_out:
	err = ext2fs_write_inode_full(fs, ino, (struct ext2_inode *)&inode, sizeof(inode));

out:
	return err;
}

static int ext2_unlink_file_by_name(ext2_filsys fs, const char *path)
{
	errcode_t err;
	ext2_ino_t dir;
	char *filename = strdup(path);
	char *base_name;

	base_name = strrchr(filename, '/');
	if (base_name) 
	{
		*base_name++ = '\0';
		err = ext2fs_namei(fs, EXT2_ROOT_INO, EXT2_ROOT_INO, filename, &dir);
		if (err) 
		{
			free(filename);
			return -1;
		}
	} else
	{
		dir = EXT2_ROOT_INO;
		base_name = filename;
	}

	err = ext2fs_unlink(fs, dir, base_name, 0, 0);
	free(filename);
	return err;
}

static int ext2_rmdir_proc(ext2_ino_t dir,
		      int	entry,
		      struct ext2_dir_entry *de,
		      int	offset,
		      int	blocksize,
		      char	*buf,
		      void	*private)
{
	struct ext2_rd_struct *rds = (struct ext2_rd_struct *) private;

	if (de->inode == 0)
		return 0;
	
	int len = ext2fs_dirent_name_len(de);
	if (len == 1 && de->name[0] == '.')
		return 0;

	if (len == 2 && de->name[0] == '.' && de->name[1] == '.') 
		rds->parent = de->inode;
	else
		rds->empty = 0;
	return 0;
}

static int ext2_fs_remove(void *vfs, const char *path)
{
	fs_t *fs = (fs_t *)vfs;
	ext2_filsys e2fs = (ext2_filsys)fs->data;
	ext2_ino_t ino = 0;
	struct ext2_inode inode;
	int err;
	
	err = ext2fs_namei(e2fs, EXT2_ROOT_INO, EXT2_ROOT_INO, path, &ino);
	if (err || ino == 0)
		goto out;
	err = ext2fs_read_inode(e2fs, ino, &inode);
	if (err)
		goto out;

	if (LINUX_S_ISDIR(inode.i_mode))
	{
		struct ext2_rd_struct rds;
		rds.parent = 0;
		rds.empty = 1;

		err = ext2fs_dir_iterate2(e2fs, ino, 0, 0, ext2_rmdir_proc, &rds);
		if (err)
			goto out;
		
		if (rds.empty == 0)
			return -ENOTEMPTY;

		err = ext2_unlink_file_by_name(e2fs, path);
		if (err)
			goto out;

		/* Directories have to be "removed" twice. */
		err = ext2_remove_inode(e2fs, ino);
		if (err)
			goto out;
		err = ext2_remove_inode(e2fs, ino);
		if (err)
			goto out;

		if (rds.parent)
		{
			err = ext2fs_read_inode_full(e2fs, rds.parent,
							 (struct ext2_inode *)&inode,
							 sizeof(inode));
			if (err)
				goto out;

			if (inode.i_links_count > 1)
				inode.i_links_count--;

			err = ext2fs_write_inode_full(e2fs, rds.parent,
							  (struct ext2_inode *)&inode,
							  sizeof(inode));
		}
	} else
	{
		err = ext2_unlink_file_by_name(e2fs, path);
		if (!err)
			err = ext2_remove_inode(e2fs, ino);
	}

out:
	ext2fs_flush(e2fs);
	return ext2_error(err);
}

static int ext2_update_dotdot(ext2_ino_t dir,
				int entry, struct ext2_dir_entry *dirent,
				int offset, int blocksize,
				char *buf, void *priv_data)
{
	struct ext2_upd_dotdot *ud = priv_data;

	if (ext2fs_dirent_name_len(dirent) == 2 &&
	    dirent->name[0] == '.' && dirent->name[1] == '.') {
		dirent->inode = ud->new_dotdot;
		return DIRENT_CHANGED | DIRENT_ABORT;
	}

	return 0;
}

static int ext2_fs_rename(void *vfs, const char *from, const char *to)
{
	fs_t *fs = (fs_t *)vfs;
	ext2_filsys e2fs = (ext2_filsys)fs->data;
	ext2_ino_t from_ino, to_ino, to_dir_ino, from_dir_ino;
	struct ext2_inode inode;
	char *temp_to = NULL, *temp_from = NULL;
	char *cp, a;
	struct ext2_upd_dotdot ud;
	int err;

	err = ext2fs_namei(e2fs, EXT2_ROOT_INO, EXT2_ROOT_INO, from, &from_ino);
	if (err || from_ino == 0)
		return -ENOENT;

	err = ext2fs_namei(e2fs, EXT2_ROOT_INO, EXT2_ROOT_INO, to, &to_ino);
	if (err)
	{
		if (err == EXT2_ET_FILE_NOT_FOUND)
			to_ino = 0;
		else
			return ext2_error(err);
	}

	/* Already the same file? */
	if (to_ino == from_ino)
		return 0;

	temp_to = strdup(to);
	if (!temp_to)
	{
		err = -ENOMEM;
		goto out;
	}

	temp_from = strdup(from);
	if (!temp_from)
	{
		err = -ENOMEM;
		goto out;
	}

	/* Find parent dir of the source and check write access */
	cp = strrchr(temp_from, '/');
	if (!cp)
	{
		err = -EINVAL;
		goto out;
	}

	a = *(cp + 1);
	*(cp + 1) = 0;
	err = ext2fs_namei(e2fs, EXT2_ROOT_INO, EXT2_ROOT_INO, temp_from, &from_dir_ino);
	*(cp + 1) = a;
	if (err)
		goto out;
	if (from_dir_ino == 0)
	{
		err = -ENOENT;
		goto out;
	}

	/* If the target exists, unlink it first */
	if (to_ino != 0)
	{
		err = ext2fs_read_inode(e2fs, to_ino, &inode);
		if (err)
			goto out;

		if (!LINUX_S_ISDIR(inode.i_mode))
		{
			err = -ENOTDIR;
			goto out;
		}

		to_dir_ino = to_ino;
	} else
	{
		/* Find parent dir of the destination and check write access */
		cp = strrchr(temp_to, '/');
		if (!cp)
		{
			err = -EINVAL;
			goto out;
		}

		a = *(cp + 1);
		*(cp + 1) = 0;
		err = ext2fs_namei(e2fs, EXT2_ROOT_INO, EXT2_ROOT_INO, temp_to,
				   &to_dir_ino);
		*(cp + 1) = a;
		if (err)
			goto out;
		if (to_dir_ino == 0)
		{
			err = -ENOENT;
			goto out;
		}
	}

	/* Get ready to do the move */
	err = ext2fs_read_inode(e2fs, from_ino, &inode);
	if (err)
		goto out;

	/* Link in the new file */
	err = ext2fs_link(e2fs, to_dir_ino, cp + 1, from_ino, ext2_file_type(inode.i_mode));
	if (err == EXT2_ET_DIR_NO_SPACE) 
	{
		err = ext2fs_expand_dir(e2fs, to_dir_ino);
		if (err)
			goto out;

		err = ext2fs_link(e2fs, to_dir_ino, cp + 1, from_ino, ext2_file_type(inode.i_mode));
	}
	if (err)
		goto out;

	/* Update '..' pointer if dir */
	err = ext2fs_read_inode(e2fs, from_ino, &inode);
	if (err)
		goto out;

	if (LINUX_S_ISDIR(inode.i_mode)) 
	{
		ud.new_dotdot = to_dir_ino;
		err = ext2fs_dir_iterate2(e2fs, from_ino, 0, NULL, ext2_update_dotdot, &ud);
		if (err)
			goto out;

		/* Decrease from_dir_ino's links_count */
		err = ext2fs_read_inode(e2fs, from_dir_ino, &inode);
		if (err)
			goto out;

		inode.i_links_count--;
		err = ext2fs_write_inode(e2fs, from_dir_ino, &inode);
		if (err)
			goto out;

		/* Increase to_dir_ino's links_count */
		err = ext2fs_read_inode(e2fs, to_dir_ino, &inode);
		if (err)
			goto out;

		inode.i_links_count++;
		err = ext2fs_write_inode(e2fs, to_dir_ino, &inode);
		if (err)
			goto out;
	}

	/* Remove the old file */
	err = ext2_unlink_file_by_name(e2fs, from);
	if (err)
		goto out;

out:
	ext2fs_flush(e2fs);

	if (temp_from)
		free(temp_from);
	if (temp_to)
		free(temp_to);
	return ext2_error(err);
}

static int ext2_fs_list_iter(ext2_ino_t dir, int entry,
			   struct ext2_dir_entry *de,
			   int offset, int blocksize, char *buf, void *data)
{
	struct ext2_dir_iter *iter = (struct ext2_dir_iter *)data;
	char namebuf[EXT2_NAME_LEN + 1];

	int len = ext2fs_dirent_name_len(de);
	if ((len == 1 && de->name[0] == '.') ||
		(len == 2 && de->name[0] == '.' && de->name[1] == '.'))
		return 0;

	struct ext2_inode inode;
	int err = ext2fs_read_inode(iter->e2fs, de->inode, &inode);
	if (err)
		return 0;

	memcpy(namebuf, de->name, len);
	namebuf[len] = 0;
	iter->filter(iter->ctx, namebuf, inode.i_size, inode.i_mtime,
		LINUX_S_ISDIR(inode.i_mode)?F_FLAG_DIR:0);
	return 0;
}

static int ext2_fs_list(void *vfs, void *ctx, const char *path, _filter filter)
{
	fs_t *fs = (fs_t *)vfs;
	ext2_filsys e2fs = (ext2_filsys)fs->data;
	struct ext2_dir_iter iter;
	ext2_ino_t ino = 0;

	errcode_t err = ext2fs_namei(e2fs, EXT2_ROOT_INO, EXT2_ROOT_INO, path, &ino);
	if (err)
		return ext2_error(err);
	else if (ino == 0)
		return -1;

	iter.ctx = ctx;
	iter.e2fs = e2fs;
	iter.filter = filter;
	
	err = ext2fs_dir_iterate2(e2fs, ino, 0, NULL, ext2_fs_list_iter, &iter);
	return ext2_error(err);
}

static int ext2_fs_stat(void *vfs, const char *path, stat_t *stat)
{
	fs_t *fs = (fs_t *)vfs;
	ext2_filsys e2fs = (ext2_filsys)fs->data;
	ext2_ino_t ino = 0;
	struct ext2_inode inode;
	int err;

	err = ext2fs_namei(e2fs, EXT2_ROOT_INO, EXT2_ROOT_INO, path, &ino);
	if (err)
		return ext2_error(err);
	else if (ino == 0)
		return -1;

	err = ext2fs_read_inode(e2fs, ino, &inode);
	if (err)
		return ext2_error(err);

	stat->size = inode.i_size;
	stat->time = inode.i_mtime; 
	stat->flags = LINUX_S_ISDIR(inode.i_mode)?F_FLAG_DIR:0;
	return 0;
}

static int ext2_fs_statfs(void *vfs, stat_fs_t *stat_fs)
{
	fs_t *fs = (fs_t *)vfs;
	ext2_filsys e2fs = (ext2_filsys)fs->data;

	stat_fs->name = strndup((char*)e2fs->super->s_volume_name, EXT2_LABEL_LEN);
	stat_fs->size = (int64_t)ext2fs_blocks_count(e2fs->super) * e2fs->blocksize;
	stat_fs->free = (int64_t)ext2fs_free_blocks_count(e2fs->super) * e2fs->blocksize;
	stat_fs->flags = 0;

	return 0;
}

static void ext2_fs_fsfree(void *vfs)
{
	fs_t *fs = (fs_t *)vfs;
	ext2_filsys e2fs = (ext2_filsys)fs->data;

	if (e2fs)
	{
		if (fs->flags & EXT2_FLAG_RW) 
		{
			e2fs->super->s_state |= EXT2_VALID_FS;
			if (e2fs->super->s_error_count)
				e2fs->super->s_state |= EXT2_ERROR_FS;
			ext2fs_mark_super_dirty(e2fs);
			ext2fs_set_gdt_csum(e2fs);
			ext2fs_flush(e2fs);
		}

		ext2fs_close(e2fs);
	}

	free(fs);
}

#ifdef MODULE_FS
void* __UI fs_open(void *block, int flags)
#else
void* __UI ext2_fs_open(void *block, int flags)
#endif
{
	struct fake_name name;
	ext2_filsys e2fs = NULL;
	block_t *blk = (block_t*)block;

	strcpy(name.header, "BLK");
	name.blk = blk;

	// !!! Original will be patched for support fake_name
	int err = ext2fs_open((const char *)&name,
							(flags&FLAG_FS_OPEN_RO)?0:EXT2_FLAG_RW, 
							0, 0, unix_io_manager, &e2fs);
	if (err || !e2fs)
		goto error;

	if (!(flags&FLAG_FS_OPEN_RO))
	{
		if (ext2fs_has_feature_journal(e2fs->super))
			LOGW("Writing to the journal is not supported");
		err = ext2fs_read_inode_bitmap(e2fs);
		if (err)
			LOGE("Fail to read the inode bitmap");
		err = ext2fs_read_block_bitmap(e2fs);
		if (err)
			LOGE("Fail to read the block bitmap");
	}

	fs_t *fs = malloc(sizeof(fs_t));
	fs->open = ext2_fs_fopen;
	fs->mkdir = ext2_fs_mkdir;
	fs->rename = ext2_fs_rename;
	fs->remove = ext2_fs_remove;
	fs->list = ext2_fs_list;
	fs->stat = ext2_fs_stat;
	fs->statfs = ext2_fs_statfs;
	fs->fsfree = ext2_fs_fsfree;
	fs->data = e2fs;
	return fs;

error:
	return NULL;
}

static int ext2_error(errcode_t err)
{
	switch(err)
	{
	case 0: return 0;
	case EXT2_ET_BAD_MAGIC: return -ENFILE;
	
	case EXT2_ET_FILE_RO:
	case EXT2_ET_RO_FILSYS: return -EROFS;
	
	case EXT2_ET_EXTENT_NO_SPACE:
	case EXT2_ET_BLOCK_ALLOC_FAIL:
	case EXT2_ET_INODE_ALLOC_FAIL:
	case EXT2_ET_DIR_NO_SPACE: return -ENOSPC;
	
	case EXT2_ET_NO_DIRECTORY: return -ENOTDIR;
	
	case EXT2_ET_DB_NOT_FOUND:
	case EXT2_ET_FILE_NOT_FOUND: return -ENOENT; 
	
	case EXT2_ET_FILE_EXISTS:
	case EXT2_ET_DIR_EXISTS: return -EEXIST;
	
	case EXT2_ET_FILE_TOO_BIG: return -EFBIG;
	
	case EXT2_ET_UNIMPLEMENTED:
	case EXT2_ET_OP_NOT_SUPPORTED: return -ENOTSUP;
	
	default: return (err < INT_MIN)?-1:(int) err;	
	}
}
